@@ -7,11 +7,15 @@
 #include <algorithm>
 #include <unordered_map>
 #include <KLocalizedString>
+
+using namespace Qt::Literals::StringLiterals;
+
 class LyricsModel::LyricsModelPrivate
 {
 public:
     bool parse(const QString &lyric);
     int highlightedIndex{-1};
+    qint64 highlightedTimestamp{-1};
     bool isLRC {false};
 
     std::vector<std::pair<QString, qint64>> lyrics;
@@ -41,7 +45,7 @@ qint64 LyricsModel::LyricsModelPrivate::parseOneTimeStamp(
      * ||| || || ||
      * ||| || || |End
      * ||| || || RightBracket
-     * ||| || |Hundredths
+     * ||| || |Milliseconds
      * ||| || Period
      * ||| |Seconds
      * ||| Colon
@@ -49,12 +53,19 @@ qint64 LyricsModel::LyricsModelPrivate::parseOneTimeStamp(
      * |LeftBracket
      * Start
      * */
-    enum States {Start, LeftBracket, Minutes, Colon, Seconds, Period, Hundredths, RightBracket, End};
-    auto states {States::Start};
-    auto minute {0}, second {0}, hundred {0};
+    enum States {Start, LeftBracket, Minutes, Colon, Seconds, Period, Milliseconds, RightBracket, End};
 
-    while (begin != end) {
-        switch (begin->toLatin1()) {
+    // Keep caller iterator intact unless we successfully parse a timestamp
+    auto current = begin;
+    if (current == end || *current != '['_L1) {
+        return -1;
+    }
+
+    auto states{States::Start};
+    auto minute{0}, second{0}, ms{0}, msFactor{100};
+
+    while (current != end) {
+        switch (current->toLatin1()) {
         case '.':
             if (states == Seconds)
                 states = Period;
@@ -64,12 +75,13 @@ qint64 LyricsModel::LyricsModelPrivate::parseOneTimeStamp(
                 states = LeftBracket;
             break;
         case ']':
-            begin++;
-            if (states == Hundredths) {
-                return minute * 60 * 1000 + second * 1000 +
-                    hundred * 10; // we return milliseconds
-            }
-            else {
+            ++current;
+            if (states == Milliseconds) {
+                // we return milliseconds
+                // Commit the consumed input to caller only on success
+                begin = current;
+                return minute * 60 * 1000 + second * 1000 + ms;
+            } else {
                 return -1;
             }
         case ':':
@@ -77,44 +89,39 @@ qint64 LyricsModel::LyricsModelPrivate::parseOneTimeStamp(
                 states = Colon;
             break;
         default:
-            if (begin->isDigit()) {
+            if (current->isDigit()) {
                 switch (states) {
-                    case LeftBracket:
-                        states = Minutes;
-                        [[fallthrough]];
-                    case Minutes:
-                        minute *= 10;
-                        minute += begin->digitValue();
-                        break;
-                    case Colon:
-                        states = Seconds;
-                        [[fallthrough]];
-                    case Seconds:
-                        second *= 10;
-                        second += begin->digitValue();
-                        break;
-                    case Period:
-                        states = Hundredths;
-                        [[fallthrough]];
-                    case Hundredths:
-                        // we only parse to hundredth second
-                        if (hundred >= 100) {
-                            break;
-                        }
-                        hundred *= 10;
-                        hundred += begin->digitValue();
-                        break;
-                    default:
-                        // lyric format is corrupt
-                        break;
+                case LeftBracket:
+                    states = Minutes;
+                    [[fallthrough]];
+                case Minutes:
+                    minute *= 10;
+                    minute += current->digitValue();
+                    break;
+                case Colon:
+                    states = Seconds;
+                    [[fallthrough]];
+                case Seconds:
+                    second *= 10;
+                    second += current->digitValue();
+                    break;
+                case Period:
+                    states = Milliseconds;
+                    [[fallthrough]];
+                case Milliseconds:
+                    ms += current->digitValue() * msFactor;
+                    msFactor /= 10;
+                    break;
+                default:
+                    // lyric format is corrupt
+                    break;
                 }
             } else {
-                begin++;
                 return -1;
             }
             break;
         }
-        begin++;
+        ++current;
     }
 
     // end of lyric and no correct value found
@@ -128,17 +135,20 @@ LyricsModel::LyricsModelPrivate::parseOneLine(QString::const_iterator &begin,
     auto size{0};
     auto it = begin;
     while (begin != end) {
-        if (begin->toLatin1() != '[') {
-            size++;
-        } else
+        if (*begin == '\n'_L1) {
+            ++begin;
             break;
-        begin++;
+        }
+
+        ++begin;
+        ++size;
     }
-    if (size) {
-        return QString(--it, size); // FIXME: really weird workaround for QChar,
-                                    // otherwise first char is lost
-    } else
+
+    if (size > 0) {
+        return QString(it, size);
+    } else {
         return {};
+    }
 }
 
 /*
@@ -148,12 +158,6 @@ LyricsModel::LyricsModelPrivate::parseOneLine(QString::const_iterator &begin,
  */
 QString LyricsModel::LyricsModelPrivate::parseTags(QString::const_iterator &begin, QString::const_iterator end)
 {
-    static auto skipTillChar = [](QString::const_iterator begin, QString::const_iterator end, char endChar) {
-        while (begin != end && begin->toLatin1() != endChar) {
-            begin++;
-        }
-        return begin;
-    };
     static std::unordered_map<QString, QString> map = {
         {QStringLiteral("ar"), i18nc("@label musical artist", "Artist")},
         {QStringLiteral("al"), i18nc("@label musical album", "Album")},
@@ -165,45 +169,43 @@ QString LyricsModel::LyricsModelPrivate::parseTags(QString::const_iterator &begi
         {QStringLiteral("ve"), i18nc("@label", "Version")}};
     QString tags;
 
+    auto rollback = begin;
+
     while (begin != end) {
         // skip till tags
-        begin = skipTillChar(begin, end, '[');
+        begin = std::find(begin, end, '['_L1);
         if (begin != end) {
-            begin++;
-        }
-        else {
+            ++begin;
+        } else {
             break;
         }
 
-        auto tagIdEnd = skipTillChar(begin, end, ':');
+        auto tagIdEnd = std::find(begin, end, ':'_L1);
         auto tagId = QString(begin, std::distance(begin, tagIdEnd));
         if (tagIdEnd != end &&
             (map.count(tagId) || tagId == QStringLiteral("offset"))) {
-          tagIdEnd++;
+            ++tagIdEnd;
 
-          auto tagContentEnd = skipTillChar(tagIdEnd, end, ']');
-          bool ok = true;
-          if (map.count(tagId)) {
-            tags += i18nc(
-                "@label this is a key => value map", "%1: %2\n", map[tagId],
-                QString(tagIdEnd, std::distance(tagIdEnd, tagContentEnd)));
-          } else {
-            // offset tag
-            offset = QString(tagIdEnd, std::distance(tagIdEnd, tagContentEnd))
-                         .toLongLong(&ok);
-          }
+            auto tagContentEnd = std::find(tagIdEnd, end, ']'_L1);
+            bool ok = true;
+            if (map.count(tagId)) {
+                tags += i18nc("@label this is a key => value map", "%1: %2\n", map[tagId], QString(tagIdEnd, std::distance(tagIdEnd, tagContentEnd)));
+            } else {
+                // offset tag
+                offset = QString(tagIdEnd, std::distance(tagIdEnd, tagContentEnd)).toLongLong(&ok);
+            }
 
-          if (ok) {
-            begin = tagContentEnd;
-          } else {
-            // Invalid offset tag, we step back one to compensate the '[' we
-            // step over
-            begin--;
-            break;
-          }
+            if (ok) {
+                begin = tagContentEnd;
+                rollback = begin;
+            } else {
+                // Invalid offset: fully rollback so caller sees the whole line
+                begin = rollback;
+                break;
+            }
         } else {
-          // No tag, we step back one to compensate the '[' we step over
-          begin--;
+          // No tag, we step rollback to compensate the '[' we step over
+          begin = rollback;
           break;
         }
     }
@@ -231,7 +233,7 @@ bool LyricsModel::LyricsModelPrivate::parse(const QString &lyric)
           timeStamp = parseOneTimeStamp(begin, end);
         }
         auto string = parseOneLine(begin, end);
-        if (!string.isEmpty() && !timeStamps.empty()) {
+        if (!timeStamps.empty()) {
             for (auto time : timeStamps) {
                 lyrics.push_back({string, time});
             }
@@ -239,17 +241,16 @@ bool LyricsModel::LyricsModelPrivate::parse(const QString &lyric)
         timeStamps.clear();
     }
 
-    std::sort(lyrics.begin(),
-              lyrics.end(),
-              [](const std::pair<QString, qint64> &lhs,
-                 const std::pair<QString, qint64> &rhs) {
-                  return lhs.second < rhs.second;
-              });
+    // Keep original relative order for identical timestamps so that
+    // (original, translation) lines remain adjacent and ordered as in file.
+    std::stable_sort(lyrics.begin(), lyrics.end(), [](const std::pair<QString, qint64> &lhs, const std::pair<QString, qint64> &rhs) {
+        return lhs.second < rhs.second;
+    });
     if (offset) {
       std::transform(lyrics.begin(), lyrics.end(),
                      lyrics.begin(),
                      [this](std::pair<QString, qint64> &element) {
-                       element.second = std::max(element.second + offset, 0ll);
+                       element.second = std::max(element.second - offset, 0ll);
                        return element;
                      });
     }
@@ -284,6 +285,10 @@ QVariant LyricsModel::data(const QModelIndex &index, int role) const
         return d->lyrics.at(index.row()).first;
     case LyricsRole::TimeStamp:
         return d->lyrics.at(index.row()).second;
+    case LyricsRole::IsHighlighted: {
+        const auto &item = d->lyrics.at(index.row());
+        return (d->highlightedTimestamp >= 0 && item.second == d->highlightedTimestamp);
+    }
     }
 
     return QVariant();
@@ -300,11 +305,17 @@ void LyricsModel::setLyric(const QString &lyric)
     if (!ret && !lyric.isEmpty()) {
         d->lyrics = {{lyric, 0ll}};
         d->highlightedIndex = -1;
+        d->highlightedTimestamp = -1;
         isLRC = false;
     }
     endResetModel();
 
     Q_EMIT highlightedIndexChanged();
+    if (!d->lyrics.empty()) {
+        const auto topLeft = index(0, 0);
+        const auto bottomRight = index(d->lyrics.size() - 1, 0);
+        Q_EMIT dataChanged(topLeft, bottomRight, {LyricsRole::IsHighlighted});
+    }
     Q_EMIT lyricChanged();
     if (isLRC != d->isLRC) {
         d->isLRC = isLRC;
@@ -318,6 +329,24 @@ void LyricsModel::setPosition(qint64 position)
         return;
     }
 
+    // get the [firstRow, lastRow] inclusive range for a given timestamp
+    auto groupRowsForTimestamp = [this](qint64 timestamp) -> std::pair<int, int> {
+        if (timestamp < 0 || d->lyrics.empty()) {
+            return {-1, -1};
+        }
+        const auto key = std::make_pair(QString(), timestamp);
+        auto comp = [](const std::pair<QString, qint64> &lhs, const std::pair<QString, qint64> &rhs) {
+            return lhs.second < rhs.second;
+        };
+        const auto range = std::equal_range(d->lyrics.begin(), d->lyrics.end(), key, comp);
+        if (range.first == range.second) {
+            return {-1, -1};
+        }
+        const int firstRow = std::distance(d->lyrics.begin(), range.first);
+        const int lastRow = std::distance(d->lyrics.begin(), range.second) - 1;
+        return {firstRow, lastRow};
+    };
+
     // do binary search
     auto result =
         std::lower_bound(d->lyrics.begin(),
@@ -326,12 +355,42 @@ void LyricsModel::setPosition(qint64 position)
                          [](const std::pair<QString, qint64> &lhs, qint64 value) {
                              return lhs.second < value;
                          });
-    if (result != d->lyrics.begin()) {
-        d->highlightedIndex = std::distance(d->lyrics.begin(), --result);
-    } else {
-        d->highlightedIndex = -1;
+
+    int highlightedIndex = -1;
+    qint64 highlightedTimestamp = -1;
+
+    if (result != d->lyrics.end() && result->second == position) {
+        // exact match -> highlight the first item at this timestamp
+        highlightedIndex = std::distance(d->lyrics.begin(), result);
+        highlightedTimestamp = result->second;
+    } else if (result != d->lyrics.begin()) {
+        // otherwise highlight the previous timestamp group (last item in that group visually,
+        // but we expose timestamp so all items with same timestamp can react)
+        --result;
+        highlightedIndex = std::distance(d->lyrics.begin(), result);
+        highlightedTimestamp = result->second;
+    }
+
+    const bool indexChanged = (highlightedIndex != d->highlightedIndex);
+    const qint64 oldTimestamp = d->highlightedTimestamp;
+    const bool timestampChanged = (highlightedTimestamp != oldTimestamp);
+    if (indexChanged) {
+        d->highlightedIndex = highlightedIndex;
+        Q_EMIT highlightedIndexChanged();
+    }
+    if (timestampChanged) {
+        const auto oldRange = groupRowsForTimestamp(oldTimestamp);
+        const auto newRange = groupRowsForTimestamp(highlightedTimestamp);
+
+        d->highlightedTimestamp = highlightedTimestamp;
+
+        if (oldRange.first != -1) {
+            Q_EMIT dataChanged(index(oldRange.first, 0), index(oldRange.second, 0), {LyricsRole::IsHighlighted});
+        }
+        if (newRange.first != -1) {
+            Q_EMIT dataChanged(index(newRange.first, 0), index(newRange.second, 0), {LyricsRole::IsHighlighted});
+        }
     }
-    Q_EMIT highlightedIndexChanged();
 }
 
 int LyricsModel::highlightedIndex() const
@@ -346,7 +405,9 @@ bool LyricsModel::isLRC() const
 
 QHash<int, QByteArray> LyricsModel::roleNames() const
 {
-    return {{LyricsRole::Lyric, QByteArrayLiteral("lyric")}, {LyricsRole::TimeStamp, QByteArrayLiteral("timestamp")}};
+    return {{LyricsRole::Lyric, QByteArrayLiteral("lyric")},
+            {LyricsRole::TimeStamp, QByteArrayLiteral("timestamp")},
+            {LyricsRole::IsHighlighted, QByteArrayLiteral("isHighlighted")}};
 }
 
 #include "moc_lyricsmodel.cpp"
