@@ -15,6 +15,7 @@
 #include "modeldataloader.h"
 #include "filescanner.h"
 
+#include <QQmlEngine>
 #include <QUrl>
 #include <QAbstractListModel>
 #include <QFutureWatcher>
@@ -26,6 +27,8 @@ class ELISALIB_EXPORT TrackMetadataModel : public QAbstractListModel
 {
     Q_OBJECT
 
+    QML_ELEMENT
+
     Q_PROPERTY(QUrl coverUrl
                READ coverUrl
                NOTIFY coverUrlChanged)
@@ -48,11 +51,11 @@ class ELISALIB_EXPORT TrackMetadataModel : public QAbstractListModel
                NOTIFY lyricsChanged)
 
 public:
-
     enum ColumnRoles
     {
         ItemNameRole = Qt::UserRole + 1,
         ItemTypeRole,
+        HasDataRole,
     };
 
     Q_ENUM(ColumnRoles)
@@ -92,6 +95,7 @@ public:
     MusicListenersManager* manager() const;
 
     QString lyrics() const;
+    QString lyricsLocation() const;
 
     qulonglong databaseId() const;
 
@@ -158,7 +162,7 @@ protected:
 
     void removeDataByIndex(int index);
 
-    void addDataByName(const QString &name);
+    void addDataByIndex(int index);
 
     static QString nameFromRole(DataTypes::ColumnsRoles role);
 
@@ -194,7 +198,8 @@ private:
 
     QMutex mFileScannerMutex;
 
-    QFutureWatcher<QString> mLyricsValueWatcher;
+    // std::pair<LyricsContent, LyricsLocation>, LyricsLocation should be empty if lyrics is embedded
+    QFutureWatcher<std::pair<QString, QString>> mLyricsValueWatcher;
 };
 
 #endif // TRACKMETADATAMODEL_H
